@@ -10,20 +10,17 @@ import zlib
 
 from binascii import Error as BinasciiError
 from construct import Container, ChecksumError, CheckError
-from copy import deepcopy
-from dateutil import parser, tz
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
 from lxml import etree
 from lxml.builder import E
 from pathlib import Path
 
-from pykeepass.attachment import Attachment
-from pykeepass.entry import Entry
-from pykeepass.exceptions import *
-from pykeepass.group import Group
-from pykeepass.kdbx_parsing.kdbx import KDBX
-from pykeepass.kdbx_parsing.kdbx4 import kdf_uuids
-from pykeepass.xpath import attachment_xp, entry_xp, group_xp, path_xp
+from .attachment import Attachment
+from .entry import Entry
+from .exceptions import *
+from .group import Group
+from .kdbx_parsing import KDBX, kdf_uuids
+from .xpath import attachment_xp, entry_xp, group_xp, path_xp
 
 logger = logging.getLogger(__name__)
 
@@ -32,7 +29,6 @@ BLANK_DATABASE_FILENAME = "blank_database.kdbx"
 BLANK_DATABASE_LOCATION = os.path.join(os.path.dirname(os.path.realpath(__file__)), BLANK_DATABASE_FILENAME)
 BLANK_DATABASE_PASSWORD = "password"
 
-
 class PyKeePass():
     """Open a KeePass database
 
@@ -45,6 +41,9 @@ class PyKeePass():
             database is assumed to have no keyfile
         transformed_key (:obj:`bytes`, optional): precomputed transformed
             key.
+        decrypt (:obj:`bool`, optional): whether to decrypt XML payload.
+            Set `False` to access outer header information without decrypting
+            database.
 
     Raises:
         CredentialsError: raised when password/keyfile or transformed key
@@ -59,13 +58,14 @@ class PyKeePass():
     """
 
     def __init__(self, filename, password=None, keyfile=None,
-                 transformed_key=None):
+                 transformed_key=None, decrypt=True):
 
         self.read(
             filename=filename,
             password=password,
             keyfile=keyfile,
-            transformed_key=transformed_key
+            transformed_key=transformed_key,
+            decrypt=decrypt
         )
 
     def __enter__(self):
@@ -76,7 +76,7 @@ class PyKeePass():
         pass
 
     def read(self, filename=None, password=None, keyfile=None,
-             transformed_key=None):
+             transformed_key=None, decrypt=True):
         """
         See class docstring.
 
@@ -96,14 +96,16 @@ class PyKeePass():
                     filename,
                     password=password,
                     keyfile=keyfile,
-                    transformed_key=transformed_key
+                    transformed_key=transformed_key,
+                    decrypt=decrypt
                 )
             else:
                 self.kdbx = KDBX.parse_file(
                     filename,
                     password=password,
                     keyfile=keyfile,
-                    transformed_key=transformed_key
+                    transformed_key=transformed_key,
+                    decrypt=decrypt
                 )
 
         except CheckError as e:
@@ -154,7 +156,8 @@ class PyKeePass():
                 filename,
                 password=self.password,
                 keyfile=self.keyfile,
-                transformed_key=transformed_key
+                transformed_key=transformed_key,
+                decrypt=True
             )
         else:
             # save to temporary file to prevent database clobbering
@@ -166,7 +169,8 @@ class PyKeePass():
                     filename_tmp,
                     password=self.password,
                     keyfile=self.keyfile,
-                    transformed_key=transformed_key
+                    transformed_key=transformed_key,
+                    decrypt=True
                 )
             except Exception as e:
                 os.remove(filename_tmp)
@@ -209,10 +213,30 @@ class PyKeePass():
         and passed to `open` for faster database opening"""
         return self.kdbx.body.transformed_key
 
+    @property
+    def database_salt(self):
+       """bytes: salt of database kdf. This can be used for adding additional
+       credentials which are used in extension to current keyfile."""
+
+       if self.version == (3, 1):
+            return self.kdbx.header.value.dynamic_header.transform_seed.data
+
+       kdf_parameters = self.kdbx.header.value.dynamic_header.kdf_parameters.data.dict
+       return kdf_parameters['S'].value
+
+    @property
+    def payload(self):
+        """Encrypted payload of keepass database"""
+        # check if payload is decrypted
+        if self.kdbx.body.payload is None:
+            raise ValueError("Database is not decrypted")
+        else:
+            return self.kdbx.body.payload
+
     @property
     def tree(self):
         """lxml.etree._ElementTree: database XML payload"""
-        return self.kdbx.body.payload.xml
+        return self.payload.xml
 
     @property
     def root_group(self):
@@ -230,13 +254,13 @@ class PyKeePass():
     def groups(self):
         """:obj:`list` of :obj:`Group`: list of all Group objects in database
         """
-        return self._xpath('//Group', cast=True)
+        return self.find_groups()
 
     @property
     def entries(self):
         """:obj:`list` of :obj:`Entry`: list of all Entry objects in database,
         excluding history"""
-        return self._xpath('//Entry', cast=True)
+        return self.find_entries()
 
     def xml(self):
         """Get XML part of database as string
@@ -267,8 +291,7 @@ class PyKeePass():
                 )
             )
 
-    def _xpath(self, xpath_str, tree=None, first=False, history=False,
-               cast=False, **kwargs):
+    def _xpath(self, xpath_str, tree=None, first=False, cast=False, **kwargs):
         """Look up elements in the XML payload and return corresponding object.
 
         Internal function which searches the payload lxml ElementTree for
@@ -283,8 +306,6 @@ class PyKeePass():
             first (bool): If True, function returns first result or None.  If
                 False, function returns list of matches or empty list.  Default
                 is False.
-            history (bool): If True, history entries are included in results.
-                Default is False.
             cast (bool): If True, matches are instead instantiated as
                 pykeepass Group, Entry, or Attachment objects.  An exception
                 is raised if a match cannot be cast.  Default is False.
@@ -302,18 +323,17 @@ class PyKeePass():
 
         res = []
         for e in elements:
-            if history or e.getparent().tag != 'History':
-                if cast:
-                    if e.tag == 'Entry':
-                        res.append(Entry(element=e, kp=self))
-                    elif e.tag == 'Group':
-                        res.append(Group(element=e, kp=self))
-                    elif e.tag == 'Binary' and e.getparent().tag == 'Entry':
-                        res.append(Attachment(element=e, kp=self))
-                    else:
-                        raise Exception('Could not cast element {}'.format(e))
+            if cast:
+                if e.tag == 'Entry':
+                    res.append(Entry(element=e, kp=self))
+                elif e.tag == 'Group':
+                    res.append(Group(element=e, kp=self))
+                elif e.tag == 'Binary' and e.getparent().tag == 'Entry':
+                    res.append(Attachment(element=e, kp=self))
                 else:
-                    res.append(e)
+                    raise Exception('Could not cast element {}'.format(e))
+            else:
+                res.append(e)
 
         # return first object in list or None
         if first:
@@ -327,6 +347,9 @@ class PyKeePass():
 
         xp = ''
 
+        if not history:
+            prefix += '[not(ancestor::History)]'
+
         if path is not None:
 
             first = True
@@ -347,8 +370,7 @@ class PyKeePass():
             if tree is not None:
                 xp += '.'
 
-            if kwargs.keys():
-                xp += prefix
+            xp += prefix
 
             # handle searching custom string fields
             if 'string' in kwargs.keys():
@@ -362,8 +384,9 @@ class PyKeePass():
                 kwargs['uuid'] = base64.b64encode(kwargs['uuid'].bytes).decode('utf-8')
 
             # convert tags to semicolon separated string before building xpath
+            # FIXME: this isn't a reliable way to search tags.  e.g. searching ['tag1', 'tag2'] will match 'tag1tag2
             if 'tags' in kwargs.keys():
-                kwargs['tags'] = ';'.join(kwargs['tags'])
+                kwargs['tags'] = ' and '.join(f'contains(text(),"{t}")' for t in kwargs['tags'])
 
             # build xpath to filter results with specified attributes
             for key, value in kwargs.items():
@@ -376,7 +399,6 @@ class PyKeePass():
             xp,
             tree=tree._element if tree else None,
             first=first,
-            history=history,
             cast=True,
             **kwargs
         )
@@ -390,7 +412,7 @@ class PyKeePass():
         if recyclebin_group is None:
             return True
         uuid_str = base64.b64encode( entry_or_group.uuid.bytes).decode('utf-8')
-        elem = self._xpath('./UUID[text()="{}"]/..'.format(uuid_str), tree=recyclebin_group._element, first=True, history=False, cast=False)
+        elem = self._xpath('./UUID[text()="{}"]/..'.format(uuid_str), tree=recyclebin_group._element, first=True, cast=False)
         return elem is None
 
 
@@ -566,7 +588,7 @@ class PyKeePass():
     def binaries(self):
         if self.version >= (4, 0):
             # first byte is a prepended flag
-            binaries = [a.data[1:] for a in self.kdbx.body.payload.inner_header.binary]
+            binaries = [a.data[1:] for a in self.payload.inner_header.binary]
         else:
             binaries = []
             for elem in self._xpath('/KeePassFile/Meta/Binaries/Binary'):
@@ -593,7 +615,7 @@ class PyKeePass():
                 data = b'\x00' + data
             # add binary element to inner header
             c = Container(type='binary', data=data)
-            self.kdbx.body.payload.inner_header.binary.append(c)
+            self.payload.inner_header.binary.append(c)
         else:
             binaries = self._xpath(
                 '/KeePassFile/Meta/Binaries',
@@ -625,7 +647,7 @@ class PyKeePass():
         try:
             if self.version >= (4, 0):
                 # remove binary element from inner header
-                self.kdbx.body.payload.inner_header.binary.pop(id)
+                self.payload.inner_header.binary.pop(id)
             else:
                 # remove binary element from XML
                 binaries = self._xpath('/KeePassFile/Meta/Binaries', first=True)
@@ -692,7 +714,7 @@ class PyKeePass():
     @password.setter
     def password(self, password):
         self._password = password
-        self.credchange_date = datetime.now()
+        self.credchange_date = datetime.now(timezone.utc)
 
     @property
     def keyfile(self):
@@ -702,7 +724,7 @@ class PyKeePass():
     @keyfile.setter
     def keyfile(self, keyfile):
         self._keyfile = keyfile
-        self.credchange_date = datetime.now()
+        self.credchange_date = datetime.now(timezone.utc)
 
     @property
     def credchange_required_days(self):
@@ -739,8 +761,8 @@ class PyKeePass():
 
     @credchange_date.setter
     def credchange_date(self, date):
-        time = self._xpath('/KeePassFile/Meta/MasterKeyChanged', first=True)
-        time.text = self._encode_time(date)
+        mk_time = self._xpath('/KeePassFile/Meta/MasterKeyChanged', first=True)
+        mk_time.text = self._encode_time(date)
 
     @property
     def credchange_required(self):
@@ -748,7 +770,7 @@ class PyKeePass():
         change_date = self.credchange_date
         if change_date is None or self.credchange_required_days == -1:
             return False
-        now_date = self._datetime_to_utc(datetime.now())
+        now_date = datetime.now(timezone.utc)
         return (now_date - change_date).days > self.credchange_required_days
 
     @property
@@ -757,30 +779,23 @@ class PyKeePass():
         change_date = self.credchange_date
         if change_date is None or self.credchange_recommended_days == -1:
             return False
-        now_date = self._datetime_to_utc(datetime.now())
+        now_date = datetime.now(timezone.utc)
         return (now_date - change_date).days > self.credchange_recommended_days
 
     # ---------- Datetime Functions ----------
 
-    def _datetime_to_utc(self, dt):
-        """Convert naive datetimes to UTC"""
-
-        if not dt.tzinfo:
-            dt = dt.replace(tzinfo=tz.gettz())
-        return dt.astimezone(tz.gettz('UTC'))
-
     def _encode_time(self, value):
         """bytestring or plaintext string: Convert datetime to base64 or plaintext string"""
 
         if self.version >= (4, 0):
             diff_seconds = int(
                 (
-                    self._datetime_to_utc(value) -
+                    value -
                     datetime(
                         year=1,
                         month=1,
                         day=1,
-                        tzinfo=tz.gettz('UTC')
+                        tzinfo=timezone.utc
                     )
                 ).total_seconds()
             )
@@ -788,7 +803,7 @@ class PyKeePass():
                 struct.pack('<Q', diff_seconds)
             ).decode('utf-8')
         else:
-            return self._datetime_to_utc(value).isoformat()
+            return value.isoformat()
 
     def _decode_time(self, text):
         """datetime.datetime: Convert base64 time or plaintext time to datetime"""
@@ -797,21 +812,15 @@ class PyKeePass():
             # decode KDBX4 date from b64 format
             try:
                 return (
-                    datetime(year=1, month=1, day=1, tzinfo=tz.gettz('UTC')) +
+                    datetime(year=1, month=1, day=1, tzinfo=timezone.utc) +
                     timedelta(
                         seconds=struct.unpack('<Q', base64.b64decode(text))[0]
                     )
                 )
             except BinasciiError:
-                return parser.parse(
-                    text,
-                    tzinfos={'UTC': tz.gettz('UTC')}
-                )
+                return datetime.fromisoformat(text.replace('Z','+00:00')).replace(tzinfo=timezone.utc)
         else:
-            return parser.parse(
-                text,
-                tzinfos={'UTC': tz.gettz('UTC')}
-            )
+            return datetime.fromisoformat(text.replace('Z','+00:00')).replace(tzinfo=timezone.utc)
 
 def create_database(
         filename, password=None, keyfile=None, transformed_key=None
